package cn.com.qmth.examcloud.web.cloud;

import java.io.IOException;
import java.util.concurrent.TimeUnit;

import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cloud.client.loadbalancer.LoadBalanced;
import org.springframework.cloud.netflix.ribbon.RibbonClients;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.client.ClientHttpRequestFactory;
import org.springframework.http.client.ClientHttpResponse;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.http.client.OkHttp3ClientHttpRequestFactory;
import org.springframework.http.client.SimpleClientHttpRequestFactory;
import org.springframework.web.client.DefaultResponseErrorHandler;
import org.springframework.web.client.RestTemplate;

import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.web.bootstrap.PropertyHolder;

/**
 * cloud 配置
 *
 * @author WANGWEI
 * @date 2019年1月28日
 * @Copyright (c) 2018-2020 WANGWEI [QQ:522080330] All Rights Reserved.
 */
@Configuration
@RibbonClients(defaultConfiguration = RibbonClientsConfiguration.class)
public class CloudClientConfiguration {

	private static final ExamCloudLog DEBUG_LOG = ExamCloudLogFactory
			.getLog(CloudClientConfiguration.class);

	@Bean
	@LoadBalanced
	@Autowired
	public RestTemplate buildRestTemplate(ClientHttpRequestFactory clientHttpRequestFactory) {

		RestTemplate restTemplate = new RestTemplate();
		restTemplate.setRequestFactory(clientHttpRequestFactory);

		restTemplate.setErrorHandler(new DefaultResponseErrorHandler() {
			@Override
			public boolean hasError(ClientHttpResponse response) throws IOException {
				return false;
			}
		});

		return restTemplate;
	}

	@Bean
	public ClientHttpRequestFactory buildClientHttpRequestFactory() {

		String httpRequestFactory = PropertyHolder.getString("examcloud.rpc.httpRequestFactory",
				"OkHttp3");

		if (httpRequestFactory.equals("Simple")) {
			return buildSimpleClientHttpRequestFactory();
		} else if (httpRequestFactory.equals("HttpComponents")) {
			return buildHttpComponentsClientHttpRequestFactory();
		}
		if (httpRequestFactory.equals("OkHttp3")) {
			return buildOkHttp3ClientHttpRequestFactory();
		} else {
			DEBUG_LOG.info(
					"value of property[examcloud.rpc.httpRequestFactory] is wrong. will use default value [OkHttp3] .");
			return buildOkHttp3ClientHttpRequestFactory();
		}
	}

	/**
	 * Simple
	 *
	 * @author WANGWEI
	 * @return
	 */
	private ClientHttpRequestFactory buildSimpleClientHttpRequestFactory() {
		int connectTimeout = PropertyHolder.getInt("examcloud.rpc.connectTimeout", 2000);
		int readTimeout = PropertyHolder.getInt("examcloud.rpc.readTimeout", 60000);

		SimpleClientHttpRequestFactory clientHttpRequestFactory = new SimpleClientHttpRequestFactory();
		clientHttpRequestFactory.setReadTimeout(readTimeout);
		clientHttpRequestFactory.setConnectTimeout(connectTimeout);
		return clientHttpRequestFactory;
	}

	/**
	 * OkHttp3
	 *
	 * @author WANGWEI
	 * @return
	 */
	private ClientHttpRequestFactory buildOkHttp3ClientHttpRequestFactory() {
		int connectTimeout = PropertyHolder.getInt("examcloud.rpc.connectTimeout", 2000);
		int readTimeout = PropertyHolder.getInt("examcloud.rpc.readTimeout", 60000);
		int writeTimeout = PropertyHolder.getInt("examcloud.rpc.writeTimeout", 60000);

		OkHttp3ClientHttpRequestFactory clientHttpRequestFactory = new OkHttp3ClientHttpRequestFactory();
		clientHttpRequestFactory.setConnectTimeout(connectTimeout);
		clientHttpRequestFactory.setReadTimeout(readTimeout);
		clientHttpRequestFactory.setWriteTimeout(writeTimeout);
		return clientHttpRequestFactory;
	}

	/**
	 * HttpComponents
	 *
	 * @author WANGWEI
	 * @return
	 */
	private ClientHttpRequestFactory buildHttpComponentsClientHttpRequestFactory() {
		Boolean poolEnable = PropertyHolder.getBoolean("examcloud.rpc.pool.enable", true);

		int maxTotal = PropertyHolder.getInt("examcloud.rpc.pool.maxTotal", 200);
		long timeToLive = PropertyHolder.getLong("examcloud.rpc.pool.timeToLive", 60);
		int defaultMaxPerRoute = PropertyHolder.getInt("examcloud.rpc.pool.defaultMaxPerRoute", 10);
		int connectTimeout = PropertyHolder.getInt("examcloud.rpc.connectTimeout", 2000);
		int readTimeout = PropertyHolder.getInt("examcloud.rpc.readTimeout", 60000);

		HttpClientBuilder httpClientBuilder = HttpClientBuilder.create();

		if (poolEnable) {
			PoolingHttpClientConnectionManager cm = new PoolingHttpClientConnectionManager(
					timeToLive, TimeUnit.SECONDS);
			cm.setValidateAfterInactivity(1000);
			cm.setMaxTotal(maxTotal);
			cm.setDefaultMaxPerRoute(defaultMaxPerRoute);

			httpClientBuilder.setConnectionManager(cm);
		}

		HttpComponentsClientHttpRequestFactory clientHttpRequestFactory = new HttpComponentsClientHttpRequestFactory();
		clientHttpRequestFactory.setHttpClient(httpClientBuilder.build());
		clientHttpRequestFactory.setConnectTimeout(connectTimeout);
		clientHttpRequestFactory.setReadTimeout(readTimeout);
		return clientHttpRequestFactory;
	}

}