package cn.com.qmth.examcloud.web.interceptor;

import java.util.Map;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang3.StringUtils;
import org.springframework.http.HttpStatus;
import org.springframework.web.servlet.HandlerInterceptor;

import com.google.common.collect.Maps;
import com.google.common.util.concurrent.RateLimiter;
import com.googlecode.aviator.AviatorEvaluator;

import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.commons.util.PropertiesUtil;
import cn.com.qmth.examcloud.commons.util.Util;
import cn.com.qmth.examcloud.web.actuator.ApiStatusInfo;
import cn.com.qmth.examcloud.web.actuator.ApiStatusInfoHolder;
import cn.com.qmth.examcloud.web.bootstrap.PropertyHolder;
import cn.com.qmth.examcloud.web.enums.HttpServletRequestAttribute;
import cn.com.qmth.examcloud.web.support.ApiInfo;
import cn.com.qmth.examcloud.web.support.ServletUtil;
import cn.com.qmth.examcloud.web.support.StatusResponse;

/**
 * API 流量限制截器
 *
 * @author WANGWEI
 * @date 2019年7月24日
 * @Copyright (c) 2018-? http://qmth.com.cn All Rights Reserved.
 */
public class ApiFlowLimitedInterceptor implements HandlerInterceptor {

	private static final ExamCloudLog LOG = ExamCloudLogFactory
			.getLog(ApiFlowLimitedInterceptor.class);

	private static final RateLimiter rateLimiter;

	private static Properties props = new Properties();

	static {

		double permitsPerSecond = PropertyHolder.getInt("examcloud.api.permitsPerSecond", 100000);
		rateLimiter = RateLimiter.create(permitsPerSecond);

		new Thread(new Runnable() {
			@Override
			public void run() {

				while (true) {
					try {
						Properties newProps = new Properties();

						PropertiesUtil.loadFromResource("limited.properties", newProps);

						props = newProps;

					} catch (Exception e) {
						LOG.error("fail to refresh API config.", e);
					}
					Util.sleep(10);
				}
			}
		}).start();

	}

	@Override
	public boolean preHandle(HttpServletRequest request, HttpServletResponse response,
			Object handler) throws Exception {

		boolean enable = PropertyHolder.getBoolean("examcloud.api.flowLimited.enable", true);

		if (!enable) {
			return true;
		}

		boolean acquired = rateLimiter.tryAcquire();
		if (!acquired) {
			response.setStatus(HttpStatus.SERVICE_UNAVAILABLE.value());
			ServletUtil.returnJson(new StatusResponse("503", "limited"), response);
			return false;
		}

		ApiInfo apiInfo = (ApiInfo) request
				.getAttribute(HttpServletRequestAttribute.$_API_INFO.name());

		if (null == apiInfo) {
			return true;
		}

		ApiStatusInfo apiStatusInfo = null;
		try {
			apiStatusInfo = ApiStatusInfoHolder.getApiStatusInfo(apiInfo.getMapping());
		} catch (Exception e) {
			// ignore
		}

		if (null == apiStatusInfo) {
			return true;
		}

		String expression = null;
		try {
			expression = (String) props.get(apiInfo.getMapping());
		} catch (Exception e) {
			// ignore
		}
		if (StringUtils.isBlank(expression)) {
			return true;
		}
		expression = expression.trim();

		Map<String, Object> env = Maps.newHashMap();
		env.put("mean", apiStatusInfo.getMean());
		env.put("max", apiStatusInfo.getMax());
		env.put("min", apiStatusInfo.getMin());
		env.put("meanRate", apiStatusInfo.getMeanRate());
		env.put("oneMinRate", apiStatusInfo.getOneMinuteRate());
		env.put("errPercent", apiStatusInfo.getErrorPercent());
		env.put("errMeanRate", apiStatusInfo.getErrorMeanRate());
		boolean limited = false;
		try {
			limited = (Boolean) AviatorEvaluator.execute(expression, env, true);
		} catch (Exception e) {
			LOG.error("fail to get value of expression. expression= " + expression, e);
		}
		if (limited) {
			// 我也不知道这是啥玩意儿 by wangwei
			Double mean = apiStatusInfo.getMean();
			int sleep = mean > 1000d ? 1000 : mean.intValue();
			sleep = sleep < 100 ? 100 : sleep;
			Util.sleep(TimeUnit.MILLISECONDS, sleep);

			response.setStatus(HttpStatus.SERVICE_UNAVAILABLE.value());
			ServletUtil.returnJson(new StatusResponse("503", "limited"), response);
			return false;
		}
		return true;
	}

	@Override
	public void afterCompletion(HttpServletRequest request, HttpServletResponse response,
			Object handler, Exception ex) throws Exception {

	}
}
