package cn.com.qmth.examcloud.web.interceptor;

import java.util.Enumeration;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.ThreadContext;
import org.springframework.http.HttpStatus;
import org.springframework.web.method.HandlerMethod;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.ModelAndView;

import com.google.common.collect.Maps;

import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.commons.util.StringUtil;
import cn.com.qmth.examcloud.commons.util.ThreadLocalUtil;
import cn.com.qmth.examcloud.web.enums.HttpServletRequestAttribute;
import cn.com.qmth.examcloud.web.support.ApiId;
import cn.com.qmth.examcloud.web.support.ApiInfo;
import cn.com.qmth.examcloud.web.support.ApiInfoHolder;
import cn.com.qmth.examcloud.web.support.ServletUtil;
import cn.com.qmth.examcloud.web.support.StatusResponse;

/**
 * 首发拦截器
 *
 * @author WANGWEI
 * @date 2018年10月22日
 * @Copyright (c) 2018-2020 WANGWEI [QQ:522080330] All Rights Reserved.
 */
public class FirstInterceptor implements HandlerInterceptor {

	private static final ExamCloudLog LOG = ExamCloudLogFactory.getLog(FirstInterceptor.class);

	private static final ExamCloudLog INTERFACE_LOG = ExamCloudLogFactory
			.getLog("INTERFACE_LOGGER");

	@Override
	public boolean preHandle(HttpServletRequest request, HttpServletResponse response,
			Object handler) throws Exception {

		String traceId = request.getHeader("Trace-Id");
		if (StringUtils.isBlank(traceId)) {
			traceId = ThreadLocalUtil.next();
		} else {
			ThreadLocalUtil.setTraceId(traceId);
		}

		// 设置log4j线程上下文
		ThreadContext.put("TRACE_ID", traceId);
		ThreadContext.put("CALLER", "$$$$$");

		String path = request.getServletPath();
		String method = request.getMethod();

		Set<String> headerNames = new HashSet<String>(20);
		if (INTERFACE_LOG.isDebugEnabled()) {
			Map<String, String> headers = Maps.newHashMap();
			Enumeration<String> e = request.getHeaderNames();
			while (e.hasMoreElements()) {
				String name = (String) e.nextElement();
				String value = request.getHeader(name);
				headers.put(name, value);
				headerNames.add(name);
			}
			INTERFACE_LOG.debug(StringUtil.join("[preHandle]. path=\"", path, "\"; method=[",
					method, "]", "; headers=", JsonUtil.toJson(headers)));
		}

		if (path.equals("/error")) {
			response.setStatus(HttpStatus.NOT_FOUND.value());
			ServletUtil.returnJson(new StatusResponse(String.valueOf(HttpStatus.NOT_FOUND.value()),
					String.valueOf(HttpStatus.NOT_FOUND.getReasonPhrase())), response);
			return false;
		}

		if ((!org.springframework.http.HttpMethod.OPTIONS.matches(method))
				&& handler instanceof HandlerMethod) {
			HandlerMethod handlerMethod = (HandlerMethod) handler;
			ApiId apiId = handlerMethod.getMethodAnnotation(ApiId.class);

			ApiInfo apiInfo = null;
			if (null != apiId) {
				apiInfo = ApiInfoHolder.getApiInfo(apiId.value());
			} else {
				apiInfo = ApiInfoHolder.getApiInfo(handlerMethod.getMethod());
			}
			String mapping = apiInfo.getMapping();
			request.setAttribute(HttpServletRequestAttribute.$_API_INFO.name(), apiInfo);

			if (INTERFACE_LOG.isDebugEnabled()) {
				INTERFACE_LOG.debug("[preHandle]. mapping = " + mapping);
			}
			request.setAttribute(HttpServletRequestAttribute.$_MAPPING.name(), mapping);

		} else {
			if (INTERFACE_LOG.isDebugEnabled()) {
				INTERFACE_LOG.debug("not handle by HandlerMethod.");
			}

			String mapping = StringUtils.join("_[", path, "][", method, "]");

			if (INTERFACE_LOG.isDebugEnabled()) {
				INTERFACE_LOG.debug("[preHandle]. mapping = " + mapping);
			}

			request.setAttribute(HttpServletRequestAttribute.$_MAPPING.name(), mapping);
			return true;
		}

		return true;
	}

	@Override
	public void postHandle(HttpServletRequest request, HttpServletResponse response, Object handler,
			ModelAndView modelAndView) throws Exception {
		LOG.debug("postHandle... ...");
	}

	@Override
	public void afterCompletion(HttpServletRequest request, HttpServletResponse response,
			Object handler, Exception ex) throws Exception {
		LOG.debug("afterCompletion... ...");
		// 清理log4j线程上下文
		ThreadContext.clearAll();
	}

}