package cn.com.qmth.examcloud.web.security;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.ThreadContext;
import org.springframework.http.HttpStatus;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.ModelAndView;

import cn.com.qmth.examcloud.api.commons.CloudService;
import cn.com.qmth.examcloud.api.commons.security.bean.AccessApp;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.commons.util.ByteUtil;
import cn.com.qmth.examcloud.commons.util.SHA256;
import cn.com.qmth.examcloud.commons.util.StringUtil;
import cn.com.qmth.examcloud.web.enums.HttpServletRequestAttribute;
import cn.com.qmth.examcloud.web.support.ApiInfo;
import cn.com.qmth.examcloud.web.support.ServletUtil;
import cn.com.qmth.examcloud.web.support.StatusResponse;

/**
 * spring cloud 请求接入
 *
 * @author WANGWEI
 * @date 2018年5月22日
 * @Copyright (c) 2018-? http://qmth.com.cn All Rights Reserved.
 */
public final class RpcInterceptor implements HandlerInterceptor {

	private static final ExamCloudLog LOG = ExamCloudLogFactory.getLog(RpcInterceptor.class);

	private ResourceManager resourceManager;

	/**
	 * 构造函数
	 *
	 * @param resourceManager
	 */
	public RpcInterceptor(ResourceManager resourceManager) {
		super();
		this.resourceManager = resourceManager;
	}

	@Override
	public boolean preHandle(HttpServletRequest request, HttpServletResponse response,
			Object handler) throws Exception {
		LOG.debug("preHandle... ...");

		ApiInfo apiInfo = (ApiInfo) request
				.getAttribute(HttpServletRequestAttribute.$_API_INFO.name());
		if (null == apiInfo) {
			return true;
		}
		Class<?> ctrClass = apiInfo.getBeanType();
		if (!CloudService.class.isAssignableFrom(ctrClass)) {
			return true;
		}

		String appId = request.getHeader("App-Id");
		String appCode = request.getHeader("App-Code");
		String timestamp = request.getHeader("timestamp");
		String accessToken = request.getHeader("Access-Token");
		if (StringUtils.isBlank(appId)) {
			// 401
			response.setStatus(HttpStatus.UNAUTHORIZED.value());
			ServletUtil
					.returnJson(new StatusResponse(String.valueOf(HttpStatus.UNAUTHORIZED.value()),
							"'App-Id' is blank"), response);
			return false;
		}
		Long appIdLong = null;
		try {
			appIdLong = Long.parseLong(appId);
		} catch (Exception e) {
			// 401
			response.setStatus(HttpStatus.UNAUTHORIZED.value());
			ServletUtil
					.returnJson(new StatusResponse(String.valueOf(HttpStatus.UNAUTHORIZED.value()),
							"'App-Id' must be a long"), response);
			return false;
		}
		if (StringUtils.isBlank(appCode)) {
			// 401
			response.setStatus(HttpStatus.UNAUTHORIZED.value());
			ServletUtil
					.returnJson(new StatusResponse(String.valueOf(HttpStatus.UNAUTHORIZED.value()),
							"'App-Code' is blank"), response);
			return false;
		}
		if (StringUtils.isBlank(timestamp)) {
			// 401
			response.setStatus(HttpStatus.UNAUTHORIZED.value());
			ServletUtil
					.returnJson(new StatusResponse(String.valueOf(HttpStatus.UNAUTHORIZED.value()),
							"'App-Code' is blank"), response);
			return false;
		}
		Long timestampLong = null;
		try {
			timestampLong = Long.parseLong(timestamp);
		} catch (Exception e) {
			// 401
			response.setStatus(HttpStatus.UNAUTHORIZED.value());
			ServletUtil
					.returnJson(new StatusResponse(String.valueOf(HttpStatus.UNAUTHORIZED.value()),
							"'timestamp' must be a long"), response);
			return false;
		}

		if (StringUtils.isBlank(accessToken)) {
			// 401
			response.setStatus(HttpStatus.UNAUTHORIZED.value());
			ServletUtil
					.returnJson(new StatusResponse(String.valueOf(HttpStatus.UNAUTHORIZED.value()),
							"'Access-Token' is blank"), response);
			return false;
		}

		AccessApp accessApp = resourceManager.getAccessApp(appIdLong);
		if (null == accessApp) {
			// 401
			response.setStatus(HttpStatus.UNAUTHORIZED.value());
			ServletUtil
					.returnJson(new StatusResponse(String.valueOf(HttpStatus.UNAUTHORIZED.value()),
							"'App-Id' is wrong"), response);
			return false;
		}
		if (!appCode.equals(accessApp.getAppCode())) {
			// 401
			response.setStatus(HttpStatus.UNAUTHORIZED.value());
			ServletUtil
					.returnJson(new StatusResponse(String.valueOf(HttpStatus.UNAUTHORIZED.value()),
							"'App-Id' & 'App-Code' are wrong"), response);
			return false;
		}

		if (null != accessApp.getTimeRange()) {
			long currentTimeMillis = System.currentTimeMillis();
			if (Math.abs(currentTimeMillis - timestampLong) > accessApp.getTimeRange()) {
				response.setStatus(HttpStatus.UNAUTHORIZED.value());
				ServletUtil.returnJson(
						new StatusResponse(String.valueOf(HttpStatus.UNAUTHORIZED.value()),
								"'timestamp' is out"),
						response);
				return false;
			}
		}

		String secretKey = accessApp.getSecretKey();
		String joinStr = StringUtil.join(appId, appCode, timestamp, secretKey);
		byte[] bytes = SHA256.encode(joinStr);
		String hexAscii = ByteUtil.toHexAscii(bytes);

		if (!hexAscii.equalsIgnoreCase(accessToken)) {
			response.setStatus(HttpStatus.UNAUTHORIZED.value());
			ServletUtil.returnJson(new StatusResponse(
					String.valueOf(HttpStatus.UNAUTHORIZED.value()), "access failure"), response);
			return false;
		}

		ThreadContext.put("CALLER", "APP:" + accessApp.getAppCode());
		return true;
	}

	@Override
	public void postHandle(HttpServletRequest request, HttpServletResponse response, Object handler,
			ModelAndView modelAndView) throws Exception {
		LOG.debug("postHandle... ...");
	}

	@Override
	public void afterCompletion(HttpServletRequest request, HttpServletResponse response,
			Object handler, Exception ex) throws Exception {
		LOG.debug("afterCompletion... ...");
	}

	public void setResourceManager(ResourceManager resourceManager) {
		this.resourceManager = resourceManager;
	}

}
